package sensor

import (
	"fmt"
	"time"

	"github.com/MichaelS11/go-dht"
	uuid "github.com/satori/go.uuid"
	"github.com/volker-raschek/flucky/pkg/types"
)

// DHT22 is a sensor to measure humidity and temperature.
type DHT22 struct {
	*types.Sensor
}

// GetEnabled returns a boolen if the sensor is enabled or disabled
func (s *DHT22) GetEnabled() bool {
	return s.SensorEnabled
}

// GetSensorModel returns the sensor model
func (s *DHT22) GetSensorModel() types.SensorModel {
	return s.Sensor.SensorModel
}

// ReadHumidity measure the humidity
func (s *DHT22) ReadHumidity() (*types.Humidity, error) {
	err := dht.HostInit()
	if err != nil {
		return nil, fmt.Errorf("HostInit error: %v", err)
	}

	gpio, err := types.GPIOToString(*s.GPIONumber)
	if err != nil {
		return nil, err
	}

	dht, err := dht.NewDHT(gpio, dht.Celsius, "")
	if err != nil {
		return nil, fmt.Errorf("NewDHT error: %v", err)
	}

	humidityValue, _, err := dht.Read()
	if err != nil {
		return nil, fmt.Errorf("Read error: %v", err)
	}

	humidity := &types.Humidity{
		HumidityID:       uuid.NewV4().String(),
		HumidityValue:    humidityValue,
		HumidityFromDate: time.Now(),
		HumidityTillDate: time.Now(),
		SensorID:         s.SensorID,
	}

	return humidity, nil
}

// ReadTemperature measure the temperature
func (s *DHT22) ReadTemperature() (*types.Temperature, error) {
	err := dht.HostInit()
	if err != nil {
		return nil, fmt.Errorf("HostInit error: %v", err)
	}

	gpio, err := types.GPIOToString(*s.GPIONumber)
	if err != nil {
		return nil, err
	}

	dht, err := dht.NewDHT(gpio, dht.Celsius, "")
	if err != nil {
		return nil, fmt.Errorf("NewDHT error: %v", err)
	}

	_, temperatureValue, err := dht.Read()
	if err != nil {
		return nil, fmt.Errorf("Read error: %v", err)
	}

	temperature := &types.Temperature{
		TemperatureID:       uuid.NewV4().String(),
		TemperatureValue:    temperatureValue,
		TemperatureFromDate: time.Now(),
		TemperatureTillDate: time.Now(),
		SensorID:            s.SensorID,
	}

	return temperature, nil
}
